import contextlib

import pandas as pd


def get_engle_sentiment_index(
    columns: str = None,
    shift: int = 1,
    precision: int = 6,
    multiplying_factor: float = 100.0,
) -> pd.DataFrame:
    col_names = ["Date", "wsj", "chneg"]
    df = pd.read_excel("Engle_CCN_index_data.xlsx", usecols="A,C,E", names=col_names)
    df["Date"] = pd.to_datetime(df["Date"] * 100 + 1, format="%Y%m%d") + pd.DateOffset(
        months=shift
    )
    df["wsj"] *= multiplying_factor
    df["chneg"] *= multiplying_factor
    rounded_df = df.round(precision).dropna()
    if columns is not None:
        return rounded_df[["Date"] + [columns]].dropna()
    # interpolate missing days linear
    with contextlib.suppress(ValueError):
        rounded_df.set_index("Date", inplace=True)
        interpolated_df = rounded_df.resample("D").interpolate(method="linear")
    return interpolated_df.dropna()


def ar1_transformation(column):
    # Shift the column values by 1 to get lagged values
    lagged_column = column.shift(1)
    # Subtract lagged values from original values
    transformed_column = column - lagged_column
    # Replace NaN with 0 (for the first row)
    transformed_column.fillna(0, inplace=True)
    return transformed_column


def get_mccc_sentiment_index(
    columns="Overall", precision: int = None, shift: int = 0, ar1: bool = False
) -> pd.DataFrame:
    df = pd.read_csv("mccc_data.csv", parse_dates=["Date"])
    df["Date"] += pd.DateOffset(days=shift)
    if precision is not None:
        df = df.round(precision)
    if isinstance(columns, list):
        df = df[["Date"] + columns]
    elif isinstance(columns, str):
        df = df[["Date"] + [columns]]
    df.set_index("Date", inplace=True)
    df = df.loc[df["Overall"] != 0]
    # perform auto-regression
    if ar1:
        ar1_df = df.copy()
        for column in df.columns:
            ar1_df[column] = ar1_transformation(df[column])
    else:
        ar1_df = df
    ar1_df.rename({"Overall": "mccc"}, axis=1, inplace=True)
    return ar1_df



import numpy as np
from mfoci.response_vars.stock_return_vars import load_stock_returns


def load_volatility_index(
    ticker="^VIX", start_date="2004-01-01", end_date="2024-01-01"
):
    """
    Load VIX index data

    :return: pd.DataFrame
    """
    df = load_stock_returns(ticker, start_date=start_date, end_date=end_date)
    df[ticker] = df[ticker].apply(np.log).diff()
    df.dropna(inplace=True)
    return df



from mfoci.factors.fama_french import get_fama_french_data
from mfoci.helpers.filterer import filter_for_common_indices
from mfoci.methods import select_factors
from mfoci.response_vars.stock_return_vars import load_stock_returns
from mfoci.response_vars.volatility_vars import load_volatility_index
from mfoci.methods.multivar_chatterjee import xi_q_n_calculate
from mfoci.methods.foci import codec

__all__ = [
    "codec",
    "filter_for_common_indices",
    "get_fama_french_data",
    "load_stock_returns",
    "load_volatility_index",
    "select_factors",
    "xi_q_n_calculate",
]



import pandas as pd


def load_index_returns():
    filename = "MSCI_countries_USD_data.xlsx"
    df = pd.read_excel(filename, usecols="C,D,E")
    df.dropna(inplace=True)
    df.drop_duplicates(inplace=True)
    countries = ["AR", "BR"]  # , "CH", "ES", "GB", "IT", "SE", "ZA"]
    # ten_countries = df["ISO"].unique()[:10]
    df = df[df["ISO"].isin(countries)]
    df["Date"] = pd.to_datetime(df["Date"], format="%Y-%m-%d %H:%M:%S")
    df.set_index("Date", inplace=True)
    pivot_df = df.pivot(columns="ISO", values="Price Close")
    returns_df = pivot_df.pct_change().dropna()
    # compute correlation matrix of the columns
    corr_matrix = returns_df.corr()
    print(corr_matrix)
    return returns_df



import decimal

import numpy as np
import pandas as pd
from scipy.spatial import cKDTree
from scipy.stats import rankdata


# Helper functions for codec
# noinspection PyPep8Naming
def codec(Y, Z, X=None, na_rm=True):
    """
    The conditional dependence coefficient (CODEC) is a measure of the amount of
    conditional dependence between a random variable Y and a random vector Z given
    a random vector X, based on an i.i.d. sample of (Y, Z, X).
    The coefficient is asymptotically guaranteed to be between 0 and 1.

    This implementation translates the FOCI.codec R method to Python.

    Parameters
    ----------
    Y : array-like
        The response variable.
    Z : array-like
        The conditioning variable.
    X : array-like, optional
        The conditioning variable. If None, the unconditional CODEC is calculated.
    na_rm : bool, optional
        A boolean value indicating whether to remove NAs. The default is True.

    Returns
    -------
    float
        The conditional dependence coefficient.

    Raises
    ------
    ValueError
        If the number of rows of Y, X, and Z are not equal.
        If the number of rows with no NAs is less than 2.

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> from mfoci import codec
    >>> n = 1000
    >>> x = np.random.rand(n, 2)
    >>> y = (x[:, 0] + x[:, 1]) % 1
    >>> y_2 = np.random.rand(n)
    >>> x_1_reshaped = x[:, 1].reshape(-1, 1)
    >>> x_0_reshaped = x[:, 0].reshape(-1, 1)
    >>> codec_y_x = codec(y, x)
    >>> z = np.random.randn(n)
    >>> z_reshaped = z.reshape(-1, 1)
    >>> codec_y_z_x = codec(y, z_reshaped, x)
    >>> codec_y_z = codec(y, z_reshaped)
    """
    if X is None:
        if isinstance(Z, list):
            Z = np.array(Z)
        if len(np.shape(Z)) == 1:
            Z = Z.reshape(-1, 1)
        if not isinstance(Y, np.ndarray):
            Y = np.array(Y)
        if not isinstance(Z, np.ndarray):
            Z = np.array(Z)
        if len(Y) != Z.shape[0]:
            raise ValueError("Number of rows of Y and Z should be equal.")
        if na_rm:
            mask = np.isfinite(Y) & np.all(np.isfinite(Z), axis=1)
            Z = Z[mask, :]
            Y = Y[mask]

        if len(Y) < 2:
            raise ValueError("Number of rows with no NAs should be bigger than 1.")

        return estimateT(Y, Z)

    if not isinstance(Y, np.ndarray):
        Y = np.array(Y)
    if not isinstance(X, np.ndarray):
        X = np.array(X)
    if not isinstance(Z, np.ndarray):
        Z = np.array(Z)
    if len(Y) != X.shape[0] or len(Y) != Z.shape[0] or Z.shape[0] != X.shape[0]:
        raise ValueError("Number of rows of Y, X, and Z should be equal.")
    if na_rm:
        mask = (
            np.isfinite(Y)
            & np.all(np.isfinite(Z), axis=1)
            & np.all(np.isfinite(X), axis=1)
        )
        Z = Z[mask, :]
        Y = Y[mask]
        X = X[mask, :]

    if len(Y) < 2:
        raise ValueError("Number of rows with no NAs should be bigger than 1.")

    return estimateConditionalT(Y, Z, X)


# noinspection PyPep8Naming
def estimateConditionalQ(Y, X, Z):
    if not isinstance(X, np.ndarray):
        X = np.array(X)
    if not isinstance(Z, np.ndarray):
        Z = np.array(Z)

    n = len(Y)
    W = np.hstack((X, Z))

    nn_X = cKDTree(X).query(X, k=3)[1][:, 1]
    repeat_data = np.where(np.linalg.norm(X - X[nn_X], axis=1) == 0)[0]
    nn_index_X = handle_repeats(nn_X, repeat_data, X)

    nn_W = cKDTree(W).query(W, k=3)[1][:, 1]
    repeat_data_W = np.where(np.linalg.norm(W - W[nn_W], axis=1) == 0)[0]
    nn_index_W = handle_repeats(nn_W, repeat_data_W, W)

    R_Y = rankdata(Y, method="max")
    minimum_1 = np.minimum(R_Y, R_Y[nn_index_W])
    minimum_2 = np.minimum(R_Y, R_Y[nn_index_X])
    Q_n = np.sum(minimum_1 - minimum_2) / n**2
    return Q_n


# noinspection PyPep8Naming
def estimateConditionalS(Y, X):
    if not isinstance(X, np.ndarray):
        X = np.array(X)
    n = len(Y)

    nn_X = cKDTree(X).query(X, k=3)[1][:, 1]
    repeat_data = np.where(np.linalg.norm(X - X[nn_X], axis=1) == 0)[0]
    nn_index_X = handle_repeats(nn_X, repeat_data, X)

    R_Y = rankdata(Y, method="max")
    S_n = np.sum(R_Y - np.minimum(R_Y, R_Y[nn_index_X])) / n**2
    return S_n


# noinspection PyPep8Naming
def estimateConditionalT(Y, Z, X):
    S = estimateConditionalS(Y, X)
    if S == 0:
        return 1
    else:
        q = estimateConditionalQ(Y, X, Z)
        return q / S


# noinspection PyPep8Naming
def estimateQ(Y, X):
    # Convert X to a numpy array if it is not already
    X = np.array(X)

    n = len(Y)
    # Use cKDTree for nearest neighbor search
    tree = cKDTree(X)
    distances, nn_indices = tree.query(X, k=3)

    # Remove the first nearest neighbor for each x, which is x itself
    nn_index_X = nn_indices[:, 1]

    # Find all data points that are not unique
    repeat_data = np.where(distances[:, 1] == 0)[0]

    # Create a DataFrame to manage repeated data
    df_X = pd.DataFrame({"id": repeat_data, "group": nn_indices[repeat_data, 0]})

    # Function to select a random nearest neighbor
    def random_nn(ids):
        if len(ids) > 0:
            return np.random.choice(ids)
        return None

    df_X["rnn"] = df_X.groupby("group")["id"].transform(random_nn)
    nn_index_X[repeat_data] = df_X["rnn"].to_numpy()

    # Nearest neighbors with ties
    ties = np.where(distances[:, 1] == distances[:, 2])[0]
    ties = np.setdiff1d(ties, repeat_data)

    def helper_ties(a):
        a_point = X[a, :].reshape(1, -1)
        rest_points = np.delete(X, a, axis=0)
        rest_indices = np.delete(np.arange(len(X)), a)

        distances_to_others = np.linalg.norm(rest_points - a_point, axis=1)
        min_indices = np.where(distances_to_others == distances_to_others.min())[0]

        # Adjust indices since we removed one point
        adjusted_indices = rest_indices[min_indices]
        random_choice = np.random.choice(adjusted_indices)
        return random_choice

    if len(ties) > 0:
        tie_choice = [helper_ties(a) for a in ties]
        nn_index_X[ties] = np.array(tie_choice)

    R_Y = rankdata(Y, method="max")
    L_Y = rankdata(-Y, method="max")
    L_Y_dec = np.array([decimal.Decimal(float(val)) for val in L_Y])
    R_Y_dec = np.array([decimal.Decimal(float(val)) for val in R_Y])
    R_Y_nn = R_Y_dec[nn_index_X]  # R_Y at nearest neighbor indices
    min_values = np.minimum(R_Y_dec, R_Y_nn)  # Element-wise minimum
    L_Y_squared = L_Y_dec**2
    Q_n = np.mean(min_values - L_Y_squared / n) / n
    return float(Q_n)


# noinspection PyPep8Naming
def estimateS(Y):
    n = len(Y)
    L_Y = rankdata(-Y, method="max")
    L_Y_decimal = np.array([decimal.Decimal(float(val)) for val in L_Y])
    n_decimal = decimal.Decimal(n)

    S_n = np.sum(L_Y_decimal * (n_decimal - L_Y_decimal)) / (n_decimal**3)
    return float(S_n)


# noinspection PyPep8Naming
def estimateT(Y, X):
    S = estimateS(Y)
    if S == 0:
        return 1
    else:
        q = estimateQ(Y, X)
        return q / S


# noinspection PyPep8Naming
def handle_repeats(nn_index, repeat_data, X):
    if len(repeat_data) > 0:
        for i in repeat_data:
            unique_indices = np.where(np.linalg.norm(X[i] - X, axis=1) != 0)[0]
            nn_index[i] = np.random.choice(unique_indices)
    return nn_index


# Example usage
if __name__ == "__main__":
    n = 1000
    x = np.random.rand(n, 2)
    y = (x[:, 0] + x[:, 1]) % 1
    y_2 = np.random.rand(n)
    x_1_reshaped = x[:, 1].reshape(-1, 1)
    x_0_reshaped = x[:, 0].reshape(-1, 1)
    # print(codec(y, x_1_reshaped, x_0_reshaped))
    codec_y_x = codec(y, x)
    codec_y_2_x = codec(y_2, x)
    z = np.random.randn(n)
    z_reshaped = z.reshape(-1, 1)
    # print(codec(y, x, z_reshaped))
    # print(codec(y, z_reshaped, x))



import pathlib
import subprocess

import pandas as pd


def kfoci(factors, response_vars, r_path, **kwargs):
    """
    Kernel FOCI - implements slide 31 from Jonathan's presentation

    :param factors: pd.DataFrame
    :param response_vars: pd.DataFrame

    :return: list
    """
    # start r subprocess calling
    path = pathlib.Path(__file__).parent
    factors.to_csv(f"{path}/x.csv", index=False)
    response_vars.to_csv(f"{path}/y.csv", index=False)
    filename = str(path / "kfoci.R")
    cmd_list = [r_path, "--vanilla", filename, path]
    if "kernel" in kwargs:
        kernel = kwargs["kernel"]
    else:
        kernel = "rbfdot"
    cmd_list.append(kernel)
    subprocess.call(cmd_list, shell=True)
    selected_cols_df = pd.read_csv(f"{path}/selected_cols.csv")
    selected_cols = selected_cols_df.iloc[:, 0].tolist()
    print(f"KFOCI {kernel} selected indicators:")
    print(selected_cols)
    return selected_cols



from mfoci.methods.mfoci_func import mfoci
from mfoci.methods.kfoci import kfoci
from mfoci.methods.lasso import select_indicators_with_lasso


def select_factors(factors, response_vars, method="mfoci", **kwargs):
    """
    Select factors using a specified method

    :param factors: pd.DataFrame
    :param response_vars: pd.DataFrame
    :param method: str

    :return: list
    """
    if method == "mfoci":
        if "shuffle" not in kwargs:
            kwargs["shuffle"] = True
        return mfoci(factors, response_vars, shuffle=kwargs["shuffle"])
    elif method == "kfoci":
        if "r_path" not in kwargs:
            raise ValueError(
                "R path must be specified for kfoci method, "
                "e.g. 'C:/Program Files/R/R-4.3.3/bin/x64/Rscript'."
            )
        r_path = kwargs["r_path"]
        del kwargs["r_path"]
        return kfoci(factors, response_vars, r_path, **kwargs)
    elif method == "lasso":
        return select_indicators_with_lasso(factors, response_vars)
    else:
        raise ValueError(f"Method {method} not implemented.")



import yfinance as yf


def load_stock_returns(tickers, start_date="2013-01-01", end_date="2024-01-01"):
    """
    Load stock returns for a list of tickers from Yahoo Finance.
    """
    df = yf.download(tickers, start=start_date, end=end_date)
    cols = [col for col in df.columns if col[0] == "Adj Close"]
    stock_prices = df[cols]
    stock_prices.columns = [col[1] for col in stock_prices.columns]
    stock_returns = stock_prices.pct_change().dropna()
    return stock_returns



import numpy as np
from sklearn.linear_model import LassoCV, MultiTaskLassoCV


def select_indicators_with_lasso(factors, response_vars):
    y_univariate = response_vars.shape[1] == 1
    if y_univariate:
        response_vars = response_vars.iloc[:, 0]
        lasso = LassoCV(cv=5)
        # lasso = Lasso()
    else:
        lasso = MultiTaskLassoCV(cv=5, random_state=0, max_iter=10000)
    lasso.fit(factors.values, response_vars.values)
    coef = lasso.coef_
    if y_univariate:
        coef = np.abs(coef)
    else:
        coef = np.sum(np.abs(coef), axis=0)
    # sort ff5 columns by the sum of the absolute values of the coefficients
    sorted_coef = np.argsort(coef)[::-1]
    n_selected = np.sum(coef > 0)
    selected_cols = factors.iloc[:, sorted_coef].columns[:n_selected]
    print("Lasso results:")
    print(
        f"Predictive indicators for {', '.join(response_vars.columns)} "
        f"are (in this order) {', '.join(selected_cols)}"
    )
    print(f"Number of selected variables is {n_selected}.")
    rounded_coef = coef[sorted_coef].round(3)
    print(f"Average absolute coefficient per indicator is {rounded_coef}.\n")
    return selected_cols, coef



import logging

import pandas as pd
from tqdm import tqdm

from mfoci.methods.multivar_chatterjee import xi_q_n_calculate


log = logging.getLogger(__name__)


def mfoci(factors, response_vars, report_insignificant=False, shuffle=True):
    """
    Multivariate FOCI - implements slide 31 from Jonathan's presentation

    :param factors: pd.DataFrame
    :param response_vars: pd.DataFrame
    :param report_insignificant: bool
    :param shuffle: bool

    :return: list
    """
    q = response_vars.shape[1]
    p = factors.shape[1]
    selected_factors = []
    max_t = 0
    max_ts = []
    n_selected = 0
    log.info(
        f"Starting MFOCI factor selection with {p} factors and {q} response variables. "
        f"Depending on the number of selected factors "
        f"there can be up to {p} iterations."
    )
    y_order_range = range(q) if shuffle else range(1)
    for i in range(p):
        log.info(f"\nIteration {i + 1}:")
        t_js = []
        for j in tqdm(range(p)):
            if j in selected_factors:
                t_js.append(0)
                continue
            t_ks = []
            for k in y_order_range:  # shuffle y order to see effect on T^q
                shuffled_y = pd.concat(
                    [response_vars.iloc[:, k:], response_vars.iloc[:, :k]], axis=1
                )
                t_k = xi_q_n_calculate(
                    factors.iloc[:, selected_factors + [j]], shuffled_y
                )
                t_ks.append(t_k)
            t_j = sum(t_ks) / len(y_order_range)  # average over different orders of y
            t_js.append(t_j)
        if max(t_js) <= max_t and n_selected == 0:
            n_selected = i
            if not report_insignificant:
                break
        max_t = max(t_js)
        argmax = t_js.index(max_t)
        selected_factors.append(argmax)
        max_ts.append(max_t)

    t = [str(round(i, 3)) for i in max_ts]
    log.info("\nMFOCI results:")
    ind_str = ", ".join(factors.columns[selected_factors])
    log.info(
        f"Predictive indicators for {', '.join(response_vars.columns)}"
        f" are (in this order) {ind_str}.\n"
        f"The corresponding T's are {' '.join(t)}.\n"
        f"Number of selected variables is {n_selected}."
    )
    print("Done!")
    return selected_factors, max_ts



import numpy as np
import pandas as pd
from scipy import stats


def xi_q_n_calculate(xvec, yvec):
    """
    Calculate the T^q value for given multivariate x and multivariate y.
    Implements the formula at the bottom of slide 21 in Jonathan's presentation, see
    also 2022 Ansari, Fuchs - A Simple Extension of T, formula (4).
    """
    xvec.reset_index(drop=True, inplace=True)
    yvec.reset_index(drop=True, inplace=True)
    q = yvec.shape[1]

    t_nom = 0
    t_den = 0
    for i in range(q):
        y = yvec.iloc[:, i].values
        prev_y = yvec.iloc[:, :i]
        x = pd.concat([xvec, prev_y], axis=1).values
        t_nom += t_y_fat_x(x, y)
        if i > 0:
            t_den += t_y_fat_x(prev_y.values, y)
    t = 1 - (q - t_nom) / (q - t_den)
    return t


def t_y_fat_x(xvec, y):
    """
    Calculate the T^q value for given multivariate x and univariate y.
    Implements the formula at the bottom of slide 17 in Jonathan's presentation.
    Corresponds to 2021 Azadkia, Chatterjee - A Simple Measure of Conditional Dependence
    the formula directly before Theorem 2.2 with xvec = Z and y = Y.
    """
    n = len(y)
    yrank = stats.rankdata(y, method="ordinal")
    # compute for each y the number of values that are at least as large
    # l_k = count_at_least_as_large(y)
    l_k = n - yrank + 1

    nn = nearest_neighbor_indices(xvec)
    # nn_old = nearest_neighbor_indices_old(xvec)
    min_yrank_nn = np.minimum(yrank, yrank[nn])
    nom = np.mean(n * min_yrank_nn - l_k**2)
    # den = np.mean(l_k * yrank)
    den = (n + 1) * (n + 2) / 6
    frac = nom / den
    return frac


def count_at_least_as_large(y):
    """
    For each element in the array, count the number of elements
    that are at least as large as the element.

    Parameters
    ----------
    y : numpy.ndarray
        1D array of numerical values.

    Returns
    -------
    counts : numpy.ndarray
        Array of counts for each element.
    """
    n = len(y)
    counts = np.zeros(n, dtype=int)

    for i in range(n):
        counts[i] = np.sum(y >= y[i])

    return counts


def nearest_neighbor_indices(data):
    """
    Find the index of the nearest neighbor for each row in the given 2D
    array using the euclidean metric.

    Parameters
    ----------
    data : array-like, shape (n_samples, n_features)

    Returns
    -------
    nearest_indices : array, shape (n_samples,)
        The index of the nearest neighbor for each row.
    """
    n_samples = data.shape[0]
    nearest_indices = np.zeros(n_samples, dtype=int)

    for i in range(n_samples):
        # Compute the squared Euclidean distance from the i-th row to all other rows
        distances = np.sum((data - data[i]) ** 2, axis=1)
        distances[i] = np.inf  # Exclude self-distance
        nearest_indices[i] = np.argmin(distances)

    return nearest_indices



import logging

log = logging.getLogger(__name__)


def filter_for_common_indices(factors, response_vars):
    """
    Filter factors and response variables for common indices.
    """
    common_idx = response_vars.index.intersection(factors.index)
    filtered_response_vars = response_vars.loc[common_idx]
    filtered_factors = factors.loc[common_idx]
    n_response_dropped = response_vars.shape[0] - filtered_response_vars.shape[0]
    n_factors_dropped = factors.shape[0] - filtered_factors.shape[0]
    if n_response_dropped > 0:
        log.warning(f"Dropped {n_response_dropped} rows from response variables.")
    if n_factors_dropped > 0:
        log.warning(f"Dropped {n_factors_dropped} rows from factors.")
    log.info(f"Response variables shape: {filtered_response_vars.shape}")
    return filtered_factors, filtered_response_vars



import warnings

import numpy as np
import pandas_datareader as pdr


def get_fama_french_data(start_date="1963-1-1", end_date="2024-01-01"):
    """
    Get Fama French data from Kenneth French's website

    :return: pd.DataFrame
    """
    dataset_name = "F-F_Research_Data_5_Factors_2x3_daily"
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        ff5_data = pdr.get_data_famafrench(dataset_name, start=start_date, end=end_date)
    ff5 = ff5_data[0].reset_index().dropna()
    # set seed
    np.random.seed(1)
    ff5["PLA-Unif"] = np.random.rand(len(ff5))
    ff5["PLA-Gauss"] = np.random.randn(len(ff5))
    ff5["PLA-Exp"] = np.exp(np.random.randn(len(ff5)))
    # ff5["Date"] = ff5["Date"].dt.strftime("%Y-%m-%d")
    ff5.set_index("Date", inplace=True)
    del ff5["RF"]
    return ff5
